{
   Copyright (C) 2010-2018 MegaBrutal

   This unit is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This unit is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
}

{
   Unit: Common
   It holds some common definitions for MgSMTP, and some helper functions.
}


{$MODE DELPHI}

unit Common;

interface
uses Windows, SysUtils, DateUtils, Classes, INIFiles;

type

   TStringArray = array of string;
   TUnixTimeStamp = longint;


   TArgument = record
      Option, Value: string;
   end;


   TArgumentParser = class
      constructor Create(RawArguments: array of string; AllowedPrefixes: array of string);
      destructor Destroy; override;
   private
      Arguments: array of TArgument;
      procedure ParseArgument(Arg: string; const AllowedPrefixes: array of string);
   public
      function GetArgument(ID: integer): TArgument;
      function IsPresent(ArgumentName: string): boolean;
      function GetValue(ArgumentName: string; DefValue: string = ''): string;
      function ValidateArguments(ValidArguments: array of string): integer;
   end;


   TNamedObject = class
      constructor Create(const Name: string; Config: TINIFile; const Section: string);
   private
      FName: string;
      FAliases: TStrings;
   public
      property Name: string read FName;
      property Aliases: TStrings read FAliases;
      function IsItYourName(const Name: string): boolean; virtual;
   end;


   TMainServerConfig = class(TNamedObject)
      constructor Create(Config: TINIFile);
   private
      FPolicies, FMailbox, FRelay, FLog: boolean;
      FDatabytes: longint;
      {FTimeCorrection: integer;}
      FTimeOffset: integer;
      FTimeOffsetStr: string;
      FListenAddresses: TStrings;
   public
      function GetVersionStr: string;
      property ListenAddresses: TStrings read FListenAddresses;
      property Databytes: longint read FDatabytes;
      {property TimeCorrection: integer read FTimeCorrection;}
      property TimeOffset: integer read FTimeOffset;
      property TimeOffsetStr: string read FTimeOffsetStr;
      property Policies: boolean read FPolicies;
      property Mailbox: boolean read FMailbox;
      property Relay: boolean read FRelay;
      property Log: boolean read FLog;
      property VersionStr: string read GetVersionStr;
   end;


   TEMailFlags = word;

   TEMailProperties = class
      constructor Create;
   protected
      FSize: longint;
      FFlags: TEMailFlags;
   public
      procedure SetSize(Value: longint);
      procedure WriteFlags(Value: TEMailFlags);
      procedure SetFlag(Flag: TEMailFlags);
      function HasFlag(Flag: TEMailFlags): boolean;
      property Size: longint read FSize write SetSize;
      property Flags: TEMailFlags read FFlags write WriteFlags;
   end;


   TRecipient = record
      Address, RMsg: string;
      Data: integer;
   end;


   TIPNamePair = class
      constructor Create(const Name, IP: string);
   protected
      FName, FIP: string;
   public
      property Name: string read FName;
      property IP: string read FIP;
      function Copy: TIPNamePair;
   end;


   TEnvelope = class
      constructor Create;
      destructor Destroy; override;
   private
      FReturnPath, FRelayHost: string;
      FReturnPathSpecified: boolean;
      FRecipients: array of TRecipient;
   public
      function GetNumberOfRecipients: integer;
      function GetRecipient(Index: integer): TRecipient;
      function IsComplete: boolean;
      procedure AddRecipient(Address: string; Data: integer = 0; RMsg: string = ''); overload;
      procedure AddRecipient(Recipient: TRecipient); overload;
      procedure SetReturnPath(Address: string);
      procedure SetRecipientData(Index, Data: integer; RMsg: string = '');
      procedure SetAllRecipientData(Data: integer; RMsg: string = '');
      procedure SetRelayHost(HostName: string);
      property ReturnPath: string read FReturnPath write SetReturnPath;
      property RelayHost: string read FRelayHost write SetRelayHost;
   end;


   TEnvelopeArray = array of TEnvelope;


   function EMailUserName(EMail: string): string;
   function EMailHost(EMail: string): string;
   function CleanEMailAddress(EMail: string): string;
   function IsValidEMailAddress(EMail: string): boolean;
   function EMailTimeStamp(DateTime: TDateTime): string;
   function EMailTimeStampCorrected(DateTime: TDateTime): string;
   function StatusToStr(Status: integer): string;
   procedure AssignDeliveryStatusToSMTPCodes(Envelope: TEnvelope; TransactionComplete: boolean);

   function CleanEOLN(S: string): string;
   function GenerateRandomString(Length: integer): string;
   function GetAlphabetStr: string;
   function GetServiceCodeStr(Ctrl: dword): string;
   function GetWinMajorVersion: longword;
   function IsPrintableString(S: string): boolean;
   function UnixTimeStamp(DateTime: TDateTime): TUnixTimeStamp;
   function CmdlineToStringArray: TStringArray;
   procedure SplitParameters(S: string; var FirstPrm, Remainder: string; Separator: char = #32);

   function ReadLineFromStream(Stream: TStream): string;
   function WriteLineToStream(Stream: TStream; Line: string): boolean;


const

   { MgSMTP version: }
   VERSION_STR       = '0.9t';

   { Architecture: }
{$IFDEF CPU64}
   PLATFORM_BITS     =  64;
{$ELSE}
{$IFDEF CPU32}
   PLATFORM_BITS     =  32;
{$ENDIF}
{$ENDIF}

   { Delivery statuses: }
   DS_DELIVERED      = 1 shl 10;
   DS_DELAYED        = 1 shl 11;
   DS_PERMANENT      = 1 shl 12;
   DS_INTERNALFAIL   = 1 shl 13;
   DS_CONNECTIONFAIL = 1 shl 14;
   DS_UNEXPECTEDFAIL = 1 shl 15;
   DS_SMTPFAIL       = 1 shl 16;
   DS_SMTPREPLYMASK  = $000003FF;
   DS_ALLFLAGS       = $FFFFFFFF;

   { E-mail property flags: }
   EF_8BITMIME       =   1;

   DayNames: array[1..7] of shortstring = ('Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat');
   MonthNames: array[1..12] of shortstring = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
      'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');

   { Support for PRESHUTDOWN is not yet present in the Free Pascal library,
     therefore I define the necessary constants here. It's a temporary
     solution, I hope I won't need it for the next release of FPC. }
   SERVICE_ACCEPT_PRESHUTDOWN  = $00000100;
   SERVICE_CONTROL_PRESHUTDOWN = $0000000F;


var

   MainServerConfig: TMainServerConfig;


implementation


{ Unit-private functions/prodecures: }

function InStringArray(const S: string; const SA: array of string): boolean;
var i: integer;
begin
   i:= 0;
   while (i < Length(SA)) and (SA[i] <> S) do Inc(i);
   Result:= i < Length(SA);
end;

function MakeTimeOffsetStr(TimeOffset: integer): string;
var CorrS: string; CorrI: integer;
begin
   CorrI:= TimeOffset;
   CorrS:= IntToStr(Abs(CorrI));
   while Length(CorrS) < 4 do CorrS:= '0' + CorrS;
   if CorrI >= 0 then CorrS:= '+' + CorrS else CorrS:= '-' + CorrS;
   Result:= CorrS;
end;


{ Unit-public functions/procedures: }

function EMailUserName(EMail: string): string;
var p: integer;
begin
   p:= Length(EMail);
   while (p > 0) and (EMail[p] <> '@') do Dec(p);
   if p <> 0 then begin
      Result:= Copy(EMail, 1, p - 1);
   end
   else Result:= EMail;
end;

function EMailHost(EMail: string): string;
var p: integer;
begin
   p:= Length(EMail);
   while (p > 0) and (EMail[p] <> '@') do Dec(p);
   if (p <> 0) and (p < Length(EMail)) then begin
      Result:= Copy(EMail, p + 1, Length(EMail) - p);
   end
   else Result:= '';
end;

function CleanEMailAddress(EMail: string): string;
var po, pc, p: integer;
begin
   po:= Pos('<', EMail);
   pc:= Pos('>', EMail);
   if (po <> 0) and (pc <> 0) and (po < pc) then begin
      Result:= Copy(EMail, po + 1, pc - po - 1);
      p:= Pos(':', Result);
      if p <> 0 then
         Result:= Copy(Result, p + 1, Length(Result) - p);
   end
   else
      Result:= EMail;
end;

function IsValidEMailAddress(EMail: string): boolean;
begin
   { !!! TODO: Implement more strict checking later !!! }
   Result:= Pos('@', EMail) <> 0;
end;

function EMailTimeStamp(DateTime: TDateTime): string;
var Year, Month, Day: word;
begin
   DecodeDate(DateTime, Year, Month, Day);
   Result:= DayNames[DayOfWeek(DateTime)] + ' ' + MonthNames[Month] + ' '
      + FormatDateTime('dd hh:nn:ss', DateTime) + ' ' + IntToStr(Year);
end;

function EMailTimeStampCorrected(DateTime: TDateTime): string;
begin
   Result:= EMailTimeStamp(DateTime) + ' ' + MainServerConfig.TimeOffsetStr;
end;

function StatusToStr(Status: integer): string;
{ Returns the delivery status code in human-readable format. }
begin
   Result:= IntToStr(Status and (DS_ALLFLAGS - DS_SMTPREPLYMASK - DS_SMTPFAIL))
      + '+' + IntToStr(Status and DS_SMTPREPLYMASK);
end;

procedure AssignDeliveryStatusToSMTPCodes(Envelope: TEnvelope; TransactionComplete: boolean);
var i, code, cond, status: integer; Recipient: TRecipient;
begin
   for i:= 0 to Envelope.GetNumberOfRecipients - 1 do begin
      Recipient:= Envelope.GetRecipient(i);
      code:= Recipient.Data and DS_SMTPREPLYMASK;
      cond:= code div 100;
      case cond of
         0: status:= DS_DELAYED or DS_UNEXPECTEDFAIL;
         2: if TransactionComplete then status:= DS_DELIVERED
            else status:= DS_DELAYED or DS_UNEXPECTEDFAIL;
         4: status:= DS_DELAYED;
         5: status:= DS_PERMANENT;
         else status:= DS_PERMANENT or DS_UNEXPECTEDFAIL;
      end;
      if code <> 0 then status:= status or DS_SMTPFAIL;
      Envelope.SetRecipientData(i, code or status, Recipient.RMsg);
   end;
end;


function CleanEOLN(S: string): string;
begin
   while (Length(S) <> 0) and (S[Length(S)] in [#13, #10]) do Delete(S, Length(S), 1);
   Result:= S;
end;

function GenerateRandomString(Length: integer): string;
var What, Chrn, i: integer; Value: string;
begin
   Value:= '';
   for i:= 1 to Length do begin
     What:= Random(3);
     case What of
     0: begin Chrn:= Random(10)+48; Value:= Value + Chr(Chrn); end;
     1: begin Chrn:= Random(26)+65; Value:= Value + Chr(Chrn); end;
     2: begin Chrn:= Random(26)+97; Value:= Value + Chr(Chrn); end;
     end;
   end;
   Result:= Value;
end;

function GetAlphabetStr: string;
var i: byte;
begin
   Result:= '';
   for i:= Ord('0') to Ord('9') do Result:= Result + Chr(i);
   for i:= Ord('A') to Ord('Z') do Result:= Result + Chr(i);
end;

function GetServiceCodeStr(Ctrl: dword): string;
begin
   case Ctrl of
   SERVICE_CONTROL_STOP:         Result:= 'STOP';
   SERVICE_CONTROL_SHUTDOWN:     Result:= 'SHUTDOWN';
   SERVICE_CONTROL_PRESHUTDOWN:  Result:= 'PRESHUTDOWN';
   else                          Result:= IntToStr(Ctrl);
   end;
end;

function GetWinMajorVersion: longword;
var OSVersionInfo: TOSVersionInfo;
begin
   { Get OS version info. }
   OSVersionInfo.dwOSVersionInfoSize:= SizeOf(TOSVersionInfo);
   GetVersionEx(OSVersionInfo);
   Result:= OSVersionInfo.dwMajorVersion;
end;

function IsPrintableString(S: string): boolean;
{ Check if string contains only printable ASCII characters. }
var i: integer;
begin
   i:= 1;
   Result:= true;
   while Result and (i <= Length(S)) do begin
      Result:= (Ord(S[i]) > 31) and (Ord(S[i]) < 127);
      Inc(i);
   end;
end;

procedure SplitParameters(S: string; var FirstPrm, Remainder: string; Separator: char = #32);
var i: integer;
begin
   i:= pos(Separator, S);
   if i > 0 then begin
      FirstPrm:= Copy(S, 1, i - 1);
      Remainder:= Copy(S, i + 1, Length(S) - i);
   end
   else begin
      FirstPrm:= S;
      Remainder:= '';
   end;
end;

function CmdlineToStringArray: TStringArray;
var i: integer;
begin
   SetLength(Result, ParamCount);
   for i:= 1 to ParamCount do
      Result[i-1]:= ParamStr(i);
end;

function UnixTimeStamp(DateTime: TDateTime): TUnixTimeStamp;
begin
   {Result:= Trunc((DateTime - EncodeDate(1970, 1 ,1)) * 24 * 60 * 60);}
   Result:= DateTimeToUnix(DateTime);
end;


function ReadLineFromStream(Stream: TStream): string;
var S: string; B: char;
begin
   S:= '';
   try
      repeat
         B:= Char(Stream.ReadByte);
         if not (B in [#10, #13]) then S:= S + B;
      until (B = #10);
   finally
      Result:= S;
   end;
end;

function WriteLineToStream(Stream: TStream; Line: string): boolean;
const EOLN = #13#10;
begin
   Result:= true;
   Line:= Line + EOLN;
   try
      Stream.WriteBuffer(PChar(Line)^, Length(Line));
   except
      Result:= false;
   end;
end;


{ Object constructors/destructors: }

constructor TArgumentParser.Create(RawArguments: array of string; AllowedPrefixes: array of string);
var i: integer;
begin
   for i:= 0 to Length(RawArguments) - 1 do
      ParseArgument(RawArguments[i], AllowedPrefixes);
end;

destructor TArgumentParser.Destroy;
begin
   SetLength(Arguments, 0);
end;

constructor TNamedObject.Create(const Name: string; Config: TINIFile; const Section: string);
begin
   inherited Create;
   FName:= Name;
   FAliases:= TStringList.Create;
   FAliases.Delimiter:= ',';
   FAliases.DelimitedText:= FName + ',' + Config.ReadString(Section, 'Alias', '');
end;

constructor TMainServerConfig.Create(Config: TINIFile);
var i: integer; rawaddresslist: string; portlist: TStringList;
begin
   inherited Create(Config.ReadString('Server', 'Name', ''), Config, 'Server');
   FListenAddresses:= TStringList.Create;
   FListenAddresses.Delimiter:= ',';

   rawaddresslist:= Config.ReadString('Server', 'ListenAddress', '');
   if rawaddresslist <> '' then
      FListenAddresses.DelimitedText:= rawaddresslist
   else begin
      portlist:= TStringList.Create;
      portlist.Delimiter:= ',';
      portlist.DelimitedText:= Config.ReadString('Server', 'ListenPort', '25');
      for i:= 0 to portlist.Count - 1 do
         FListenAddresses.Add('0.0.0.0:' + portlist.Strings[i]);
      portlist.Free;
   end;

   FDatabytes:=      Config.ReadInteger('Server', 'Databytes', 1024 * 1024 * 1024);
   FTimeOffset:=     Config.ReadInteger('Server', 'TimeOffset', Config.ReadInteger('Server', 'TimeCorrection', 0) * 100);
   FTimeOffsetStr:=  MakeTimeOffsetStr(FTimeOffset);

   FPolicies:=  Config.ReadBool('Server', 'Policies',  false);
   FMailbox:=   Config.ReadBool('Server', 'Mailbox',   false);
   FRelay:=     Config.ReadBool('Server', 'Relay',     false);
   FLog:=       Config.ReadBool('Server', 'Log',       false);
end;

constructor TEMailProperties.Create;
begin
   inherited Create;
   SetSize(0);
   WriteFlags(0);
end;

constructor TIPNamePair.Create(const Name, IP: string);
begin
   FName:= Name;
   FIP:= IP;
end;

constructor TEnvelope.Create;
begin
   inherited Create;
   FReturnPath:= '';
   FReturnPathSpecified:= false;
   FRelayHost:= '';
   SetLength(FRecipients, 0);
end;

destructor TEnvelope.Destroy;
begin
   SetLength(FRecipients, 0);
   inherited Destroy;
end;


{ Object methods: }

procedure TArgumentParser.ParseArgument(Arg: string; const AllowedPrefixes: array of string);
var i, n: integer; found: boolean;
begin
   { Strip prefix if present. }
   i:= 0; found:= false;
   while ((i < Length(AllowedPrefixes)) and (not found)) do begin
      if pos(AllowedPrefixes[i], Arg) = 1 then
      begin
         Delete(Arg, 1, Length(AllowedPrefixes[i]));
         found:= true;
      end;
      Inc(i);
   end;

   n:= Length(Arguments);
   SetLength(Arguments, n + 1);
   SplitParameters(Arg, Arguments[n].Option, Arguments[n].Value, '=');
   { To be case-insensitive: }
   Arguments[n].Option:= UpCase(Arguments[n].Option);
end;

function TArgumentParser.GetArgument(ID: integer): TArgument;
begin
   { No index checking... you'd better use it return value of ValidateArguments. }
   Result:= Arguments[ID];
end;

function TArgumentParser.IsPresent(ArgumentName: string): boolean;
var i: integer;
begin
   i:= 0;
   while (i < Length(Arguments)) and (Arguments[i].Option <> UpCase(ArgumentName)) do
      Inc(i);
   Result:= i < Length(Arguments);
end;

function TArgumentParser.GetValue(ArgumentName: string; DefValue: string = ''): string;
var i: integer;
begin
   i:= 0;
   while (i < Length(Arguments)) and (Arguments[i].Option <> UpCase(ArgumentName)) do
      Inc(i);

   if i < Length(Arguments) then begin
      if Arguments[i].Value <> '' then
         Result:= Arguments[i].Value
      else
         Result:= DefValue;
   end
   else
      Result:= DefValue;
end;

function TArgumentParser.ValidateArguments(ValidArguments: array of string): integer;
{ Returns -1 if all arguments are valid. Otherwise, returns the ID of the first
  invalid parameter. }
var i: integer;
begin
   i:= 0;
   while (i < Length(Arguments)) and InStringArray(Arguments[i].Option, ValidArguments) do
      Inc(i);

   if i >= Length(Arguments) then
      Result:= -1
   else
      Result:= i;
end;


function TNamedObject.IsItYourName(const Name: string): boolean;
begin
   Result:= FAliases.IndexOf(Name) <> -1;
end;


function TMainServerConfig.GetVersionStr: string;
begin
   Result:= VERSION_STR;
end;


function TIPNamePair.Copy: TIPNamePair;
begin
   Result:= TIPNamePair.Create(Name, IP);
end;


procedure TEMailProperties.SetSize(Value: longint);
begin
   FSize:= Value;
end;

procedure TEMailProperties.WriteFlags(Value: TEMailFlags);
begin
   FFlags:= Value;
end;

procedure TEMailProperties.SetFlag(Flag: TEMailFlags);
begin
   FFlags:= FFlags or Flag;
end;

function TEMailProperties.HasFlag(Flag: TEMailFlags): boolean;
begin
   Result:= (FFlags and Flag) = Flag;
end;


function TEnvelope.GetNumberOfRecipients: integer;
begin
   Result:= Length(FRecipients);
end;

function TEnvelope.GetRecipient(Index: integer): TRecipient;
begin
   Result:= FRecipients[Index];
end;

function TEnvelope.IsComplete: boolean;
begin
   Result:= FReturnPathSpecified and (Length(FRecipients) > 0);
end;

procedure TEnvelope.AddRecipient(Address: string; Data: integer = 0; RMsg: string = '');
var i: integer;
begin
   i:= Length(FRecipients);
   SetLength(FRecipients, i + 1);
   FRecipients[i].Address:= Address;
   FRecipients[i].RMsg:= RMsg;
   FRecipients[i].Data:= Data;
end;

procedure TEnvelope.AddRecipient(Recipient: TRecipient);
var i: integer;
begin
   i:= Length(FRecipients);
   SetLength(FRecipients, i + 1);
   FRecipients[i]:= Recipient;
end;

procedure TEnvelope.SetRecipientData(Index, Data: integer; RMsg: string = '');
begin
   FRecipients[Index].RMsg:= RMsg;
   FRecipients[Index].Data:= Data;
end;

procedure TEnvelope.SetAllRecipientData(Data: integer; RMsg: string = '');
var i: integer;
begin
   for i:= 0 to Length(FRecipients) - 1 do
      SetRecipientData(i, Data, RMsg);
end;

procedure TEnvelope.SetReturnPath(Address: string);
begin
   FReturnPath:= Address;
   FReturnPathSpecified:= true;
end;

procedure TEnvelope.SetRelayHost(HostName: string);
begin
   FRelayHost:= HostName;
end;


end.
